@@ -2,7 +2,7 @@
 
 # All chem-fingerprint software is distributed with the following license:
 
-# Copyright (c) 2010-2011 Andrew Dalke Scientific, AB (Gothenburg, Sweden)
+# Copyright (c) 2010-2013 Andrew Dalke Scientific, AB (Gothenburg, Sweden)
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -24,19 +24,15 @@
 # CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
 
-__version__ = "1.0"
-__version_info = (1, 0, 0)
+__version__ = "1.1p1"
+__version_info = (1, 1, 1)
 SOFTWARE = "chemfp/" + __version__
 
 import os
-import __builtin__
 import itertools
 
 __all__ = ["open", "load_fingerprints", "read_structure_fingerprints",
-           "count_tanimoto_hits", "threshold_tanimoto_search",
-           "knearest_tanimoto_search", "Metadata", "FingerprintIterator",
-           "Fingerprints"]
-           
+           "Metadata", "FingerprintIterator", "Fingerprints"]
            
 
 class ChemFPError(Exception):
@@ -90,7 +86,7 @@ def read_structure_fingerprints(type, source=None, format=None, id_tag=None, err
     :param type: information about how to convert the input structure into a fingerprint
     :type type: string or Metadata
     :param source: The structure data source.
-    :type source: A filename (as a string), a file object, or None to read from stdin.
+    :type source: A filename (as a string), a file object, or None to read from stdin
     :param format: The file format and optional compression.
             Examples: 'smi' and 'sdf.gz'
     :type format: string, or None to autodetect based on the source
@@ -102,16 +98,15 @@ def read_structure_fingerprints(type, source=None, format=None, id_tag=None, err
     """ # ' # emacs cruft
     from . import types
     if isinstance(type, basestring):
-        metadata = Metadata(type=type)
+        metadata = None
     else:
         metadata = type
         if metadata.type is None:
             raise ValueError("Missing fingerprint type information in metadata")
-    try:
-        structure_fingerprinter = types.parse_type(metadata.type)
-    except ValueError, err:
-        raise ValueError("Cannot parse fingerprint type %r: %s" % (metadata.type, err))
-    return structure_fingerprinter.read_structure_fingerprints(metadata, source, format, id_tag, errors)
+        type = metadata.type
+
+    structure_fingerprinter = types.parse_type(type)
+    return structure_fingerprinter.read_structure_fingerprints(source, format, id_tag, errors, metadata=metadata)
     
 # Low-memory, forward-iteration, or better
 def open(source, format=None):
@@ -139,7 +134,7 @@ def open(source, format=None):
             print id, fp.encode("hex")
         
     :param source: The fingerprint source.
-    :type source: A filename string, a file object, or None.
+    :type source: A filename string, a file object, or None
     :param format: The file format and optional compression.
     :type format: string, or None
 
@@ -149,16 +144,18 @@ def open(source, format=None):
     format_name, compression = io.normalize_format(source, format)
 
     if format_name == "fps":
-        from . import readers
-        return readers.open_fps(source, format_name+compression)
+        from . import fps_io
+        return fps_io.open_fps(source, format_name+compression)
 
     if format_name == "fpb":
         raise NotImplementedError("fpb format support not implemented")
-
-    raise TypeError("Unable to determine fingerprint format type from %r" % (source,))
+    if format is None:
+        raise ValueError("Unable to determine fingerprint format type from %r" % (source,))
+    else:
+        raise ValueError("Unknown fingerprint format %r" % (format,))
 
 
-def load_fingerprints(reader, metadata=None, reorder=True):
+def load_fingerprints(reader, metadata=None, reorder=True, alignment=None):
     """Load all of the fingerprints into an in-memory FingerprintArena data structure
     
     The FingerprintArena data structure reads all of the fingerprints and
@@ -173,6 +170,13 @@ def load_fingerprints(reader, metadata=None, reorder=True):
     The loader may reorder the fingerprints for better search performance.
     To prevent ordering, use reorder=False.
 
+    The 'alignment' option specifies the alignment data alignment and
+    padding size for each fingerprint. A value of 8 means that each
+    fingerprint will start on a 8 byte alignment, and use storage
+    space which a multiple of 8 bytes long. The default value of None
+    determines the best alignment based on the fingerprint size and
+    available popcount methods.
+
     :param reader: An iterator over (id, fingerprint) pairs
     :type reader: a string, file object, or (id, fingerprint) iterator
     :param metadata: The metadata for the arena, if other than reader.metadata
@@ -180,6 +184,7 @@ def load_fingerprints(reader, metadata=None, reorder=True):
     :param reorder: Specify if fingerprints should be reordered for better performance
     :type reorder: True or False
     :returns: FingerprintArena
+    :param alignment: Alignment size (both data alignment and padding) 
     """
     if isinstance(reader, basestring):
         reader = open(reader)
@@ -189,7 +194,8 @@ def load_fingerprints(reader, metadata=None, reorder=True):
         metadata = reader.metadata
 
     from . import arena
-    return arena.fps_to_arena(reader, metadata=metadata, reorder=reorder)
+    return arena.fps_to_arena(reader, metadata=metadata, reorder=reorder,
+                              alignment=alignment)
 
 ##### High-level search interfaces
 
@@ -202,10 +208,10 @@ def count_tanimoto_hits(queries, targets, threshold=0.7, arena_size=100):
 
     Example::
 
-      queries = chemfp.open("queries.fps")
-      targets = chemfp.load_fingerprints("targets.fps.gz")
-      for (query_id, count) in chemfp.count_tanimoto_hits(queries, targets, threshold=0.9):
-          print query_id, "has", count, "neighbors with at least 0.9 similarity"
+        queries = chemfp.open("queries.fps")
+        targets = chemfp.load_fingerprints("targets.fps.gz")
+        for (query_id, count) in chemfp.count_tanimoto_hits(queries, targets, threshold=0.9):
+            print query_id, "has", count, "neighbors with at least 0.9 similarity"
 
     Internally, queries are processed in batches of size 'arena_size'.
     A small batch size uses less overall memory and has lower
@@ -227,18 +233,38 @@ def count_tanimoto_hits(queries, targets, threshold=0.7, arena_size=100):
     :returns:
        An iterator containing (query_id, score) pairs, one for each query
     """
-    if arena_size == 1:
-        for (query_id, query_fp) in queries:
-            targets.reset()
-            yield query_id, targets.tanimoto_count_arena(query_fp, threshold)
-        return
+    from . import fps_io
+    if isinstance(targets, fps_io.FPSReader):
+        from . import fps_search
+        count_hits = fps_search.count_tanimoto_hits_arena
+    else:
+        from . import search
+        count_hits = search.count_tanimoto_hits_arena
 
-    for query_arena in queries.iter_arenas(arena_size):
-        targets.reset()
-        results = targets.count_tanimoto_hits_arena(query_arena, threshold)
-        for item in zip(query_arena.ids, results):
-            yield item
+    ### Start the search now so compatibility errors are raised eagerly
+
+    # Start iterating through the subarenas, and get the first of those
+    subarenas = queries.iter_arenas(arena_size)
+    try:
+        first_query_arena = subarenas.next()
+    except StopIteration:
+        # There are no subarenas; return an empty iterator
+        return iter([])
+
+    # Get the first result, and hold on to it for the generator
+    first_counts = count_hits(first_query_arena, targets, threshold=threshold)
     
+    def count_tanimoto_hits():
+        # Return results for the first arena
+        for query_id, count in zip(first_query_arena.ids, first_counts):
+            yield query_id, count
+        # Return results for the rest of the arenas
+        for query_arena in subarenas:
+            counts = count_hits(query_arena, targets, threshold=threshold)
+            for query_id, count in zip(query_arena.ids, counts):
+                yield query_id, count
+    return count_tanimoto_hits()
+
 
 def threshold_tanimoto_search(queries, targets, threshold=0.7, arena_size=100):
     """Find all targets within 'threshold' of each query term
@@ -252,7 +278,7 @@ def threshold_tanimoto_search(queries, targets, threshold=0.7, arena_size=100):
 
       queries = chemfp.open("queries.fps")
       targets = chemfp.load_fingerprints("targets.fps.gz")
-      for (query_id, hits) in chemfp.threshold_tanimoto_search(queries, targets, threshold=0.8):
+      for (query_id, hits) in chemfp.id_threshold_tanimoto_search(queries, targets, threshold=0.8):
           print query_id, "has", len(hits), "neighbors with at least 0.8 similarity"
           non_identical = [target_id for (target_id, score) in hits if score != 1.0]
           print "  The non-identical hits are:", non_identical
@@ -278,17 +304,44 @@ def threshold_tanimoto_search(queries, targets, threshold=0.7, arena_size=100):
       An iterator containing (query_id, hits) pairs, one for each query.
       'hits' contains a list of (target_id, score) pairs.
     """
-    if arena_size == 1:
-        for (query_id, query_fp) in queries:
-            targets.reset()
-            yield query_id, targets.threshold_tanimoto_search_fp(query_fp, threshold)
-        return
-    
-    for query_arena in queries.iter_arenas(arena_size):
-        targets.reset()
-        results = targets.threshold_tanimoto_search_arena(query_arena, threshold)
-        for item in zip(query_arena.ids, results):
-            yield item
+    from . import fps_io
+    if isinstance(targets, fps_io.FPSReader):
+        from . import fps_search
+        threshold_search = fps_search.threshold_tanimoto_search_arena
+    else:
+        from . import search
+        threshold_search = search.threshold_tanimoto_search_arena
+
+    ### Start the search now so compatibility errors are raised eagerly
+
+    # Start iterating through the subarenas, and get the first of those
+    subarenas = queries.iter_arenas(arena_size)
+    try:
+        first_query_arena = subarenas.next()
+    except StopIteration:
+        # There are no subarenas; return an empty iterator
+        return iter([])
+
+    # Get the first result, and hold on to it for the generator
+    first_results = threshold_search(first_query_arena, targets, threshold=threshold)
+    ## Here's a thought; allow a 'result_order' parameter so I can do:
+    # if result_order is not None:
+    #    first_results.reorder(reorder)
+
+    def threshold_tanimoto_search():
+        # Return results for the first arena
+        for query_id, row in zip(first_query_arena.ids, first_results):
+            yield query_id, row
+        
+        for query_arena in subarenas:
+            results = threshold_search(query_arena, targets, threshold=threshold)
+            ## I would also need to do
+            #if result_order is not None:
+            #    first_results.reorder(reorder)
+                
+            for query_id, row in zip(query_arena.ids, results):
+                yield (query_id, row)
+    return threshold_tanimoto_search()
 
 def knearest_tanimoto_search(queries, targets, k=3, threshold=0.7, arena_size=100):
     """Find the 'k'-nearest targets within 'threshold' of each query term
@@ -309,7 +362,7 @@ def knearest_tanimoto_search(queries, targets, k=3, threshold=0.7, arena_size=10
       targets = chemfp.load_fingerprints("pubchem_subset.fps")
       
       # Find the 3 nearest hits with a similarity of at least 0.8
-      for (query_id, hits) in chemfp.knearest_tanimoto_search(queries, targets, k=3, threshold=0.8):
+      for (query_id, hits) in chemfp.id_knearest_tanimoto_search(queries, targets, k=3, threshold=0.8):
           print query_id, "has", len(hits), "neighbors with at least 0.8 similarity"
           if hits:
               target_id, score = hits[-1]
@@ -338,17 +391,151 @@ def knearest_tanimoto_search(queries, targets, k=3, threshold=0.7, arena_size=10
       An iterator containing (query_id, hits) pairs, one for each query.
       'hits' contains a list of (target_id, score) pairs, sorted by score.
     """
-    if arena_size == 1:
-        for (query_id, query_fp) in queries:
-            targets.reset()
-            yield query_id, targets.knearest_tanimoto_search_fp(query_fp, k, threshold)
-        return
+    from . import fps_io
+    if isinstance(targets, fps_io.FPSReader):
+        from . import fps_search
+        knearest_search = fps_search.knearest_tanimoto_search_arena
+    else:
+        from . import search
+        knearest_search = search.knearest_tanimoto_search_arena
+        
+    ### Start the search now so compatibility errors are raised eagerly
 
-    for query_arena in queries.iter_arenas(arena_size):
-        targets.reset()
-        results = targets.knearest_tanimoto_search_arena(query_arena, k, threshold)
-        for item in zip(query_arena.ids, results):
-            yield item
+    # Start iterating through the subarenas, and get the first of those
+    subarenas = queries.iter_arenas(arena_size)
+    try:
+        first_query_arena = subarenas.next()
+    except StopIteration:
+        # There are no subarenas; return an empty iterator
+        return iter([])
+
+    # Get the first result, and hold on to it for the generator
+    first_results = knearest_search(first_query_arena, targets, k=k, threshold=threshold)
+
+    def knearest_tanimoto_search():
+        # Return results for the first arena
+        for query_id, row in zip(first_query_arena.ids, first_results):
+            yield query_id, row
+
+        # and for the subarenas
+        for query_arena in subarenas:
+            results = knearest_search(query_arena, targets, k=k, threshold=threshold)
+            for query_id, row in zip(query_arena.ids, results):
+                yield (query_id, row)
+        
+    return knearest_tanimoto_search()
+
+def count_tanimoto_hits_symmetric(fingerprints, threshold=0.7):
+    """Find the number of other fingerprints within `threshold` of each fingerprint
+    
+    For each fingerprint in the `fingerprints` arena, find the number
+    of other fingerprints in the same arena which are at least
+    `threshold` similar to it. The arena must have pre-computed
+    popcounts. A fingerprint never matches itself.
+
+    This function returns an iterator of (fingerprint_id, count) pairs.
+
+    Example::
+
+      arena = chemfp.load_fingerprints("targets.fps.gz")
+      for (fp_id, count) in chemfp.count_tanimoto_hits_symmetric(arena, threshold=0.6):
+          print fp_id, "has", count, "neighbors with at least 0.6 similarity"
+    
+    :param fingerprints: The arena containing the fingerprints.
+    :type fingerprints: a FingerprintArena with precomputed popcount_indices
+    :param threshold: The minimum score threshold.
+    :type threshod: float between 0.0 and 1.0, inclusive
+    :returns:
+      An iterator of (fp_id, count) pairs, one for each fingerprint
+    """
+    from . import fps_io, search
+    if (isinstance(fingerprints, fps_io.FPSReader) or
+        not getattr(fingerprints, "popcount_indices", None)):
+        raise ValueError("`fingerprints` must be a FingerprintArena with pre-computed popcount indices")
+
+    # Start the search now so the errors are caught early
+    results = search.count_tanimoto_hits_symmetric(fingerprints, threshold)
+    def count_tanimoto_hits_symmetric_internal():
+        for id, count in zip(fingerprints.ids, results):
+            yield id, count
+    return count_tanimoto_hits_symmetric_internal()
+
+def threshold_tanimoto_search_symmetric(fingerprints, threshold=0.7):
+    """Find the other fingerprints within `threshold` of each fingerprint
+
+    For each fingerprint in the `fingerprints` arena, find the other
+    fingerprints in the same arena which hare at least `threshold`
+    similar to it. The arena must have pre-computed popcounts. A
+    fingerprint never matches itself.
+
+    This function returns an iterator of (fingerprint, SearchResult) pairs.
+    The SearchResult hit order is arbitrary.
+
+    Example::
+    
+      arena = chemfp.load_fingerprints("targets.fps.gz")
+      for (fp_id, hits) in chemfp.threshold_tanimoto_search_symmetric(arena, threshold=0.75):
+          print fp_id, "has", len(hits), "neighbors:"
+          for (other_id, score) in hits.get_ids_and_scores():
+              print "   %s  %.2f" % (other_id, score)
+
+    :param fingerprints: The arena containing the fingerprints.
+    :type fingerprints: a FingerprintArena with precomputed popcount_indices
+    :param threshold: The minimum score threshold.
+    :type threshod: float between 0.0 and 1.0, inclusive
+    :returns: An iterator of (fp_id, SearchResult) pairs, one for each fingerprint
+    """
+    from . import fps_io, search
+    if (isinstance(fingerprints, fps_io.FPSReader) or
+        not getattr(fingerprints, "popcount_indices", None)):
+        raise ValueError("`fingerprints` must be a FingerprintArena with pre-computed popcount indices")
+
+    # Start the search now so the errors are caught early
+    results = search.threshold_tanimoto_search_symmetric(fingerprints, threshold)
+    def threshold_tanimoto_search_symmetric_internal():
+        for id, hits in zip(fingerprints.ids, results):
+            yield id, hits
+    return threshold_tanimoto_search_symmetric_internal()
+
+def knearest_tanimoto_search_symmetric(fingerprints, k=3, threshold=0.7):
+    """Find the nearest `k` fingerprints within `threshold` of each fingerprint
+
+    For each fingerprint in the `fingerprints` arena, find the nearest
+    `k` fingerprints in the same arena which hare at least `threshold`
+    similar to it. The arena must have pre-computed popcounts. A
+    fingerprint never matches itself.
+
+    This function returns an iterator of (fingerprint, SearchResult) pairs.
+    The SearchResult hits are ordered from highest score to lowest, with
+    ties broken arbitrarily.
+
+    Example::
+    
+      arena = chemfp.load_fingerprints("targets.fps.gz")
+      for (fp_id, hits) in chemfp.knearest_tanimoto_search_symmetric(arena, k=5, threshold=0.5):
+          print fp_id, "has", len(hits), "neighbors, with scores", 
+          print ", ".join("%.2f" % x for x in hits.get_scores())
+
+    :param fingerprints: The arena containing the fingerprints.
+    :type fingerprints: a FingerprintArena with precomputed popcount_indices
+    :param k: The maximum number of nearest neighbors to find.
+    :type k: positive integer
+    :param threshold: The minimum score threshold.
+    :type threshod: float between 0.0 and 1.0, inclusive
+    :returns: An iterator of (fp_id, SearchResult) pairs, one for each fingerprint
+    """
+    from . import fps_io, search
+    if (isinstance(fingerprints, fps_io.FPSReader) or
+        not getattr(fingerprints, "popcount_indices", None)):
+        raise ValueError("`fingerprints` must be a FingerprintArena with pre-computed popcount indices")
+
+    # Start the search now so the errors are caught early
+    results = search.knearest_tanimoto_search_symmetric(fingerprints, k, threshold)
+    def knearest_tanimoto_search_symmetric_internal():
+        for id, hits in zip(fingerprints.ids, results):
+            yield id, hits
+    return knearest_tanimoto_search_symmetric_internal()
+        
 
 def check_fp_problems(fp, metadata):
     "This interface is not documented and may change in the future"
@@ -432,9 +619,11 @@ class Metadata(object):
                 pass
             else:
                 num_bytes = (num_bits + 7)//8
-        elif num_bits is not None:
+        elif num_bits is None:
+            num_bits = num_bytes * 8
+        else:
             if (num_bits + 7)//8 != num_bytes:
-                raise TypeError("num_bits of %d is incompatible with num_bytes of %d" %
+                raise ValueError("num_bits of %d is incompatible with num_bytes of %d" %
                                 (num_bits, num_bytes))
             
         self.num_bits = num_bits
@@ -452,7 +641,7 @@ class Metadata(object):
         self.date = date
 
     def __repr__(self):
-        return "Metadata(num_bits=%(num_bits)s, num_bytes=%(num_bytes)d, type=%(type)r, aromaticity=%(aromaticity)r, sources=%(sources)r, software=%(software)r, date=%(date)r)" % self.__dict__
+        return "Metadata(num_bits=%(num_bits)r, num_bytes=%(num_bytes)r, type=%(type)r, aromaticity=%(aromaticity)r, sources=%(sources)r, software=%(software)r, date=%(date)r)" % self.__dict__
 
     def __str__(self):
         from cStringIO import StringIO
@@ -526,6 +715,11 @@ class FingerprintReader(object):
                 break
             yield arena
 
+    def save(self, destination):
+        from . import io
+        io.write_fps1_output(self, destination, self.metadata)
+
+
 class FingerprintIterator(FingerprintReader):
     """A FingerprintReader for an iterator of (id, fingerprint) pairs
 
@@ -576,4 +770,44 @@ class Fingerprints(FingerprintReader):
         return self._id_fp_pairs[i]
 
     # Question: should I support other parts of the list API?
-    # I almost certainly want to support slice syntax ilke x[:5]
+    # I almost certainly want to support slice syntax like x[:5]
+
+def get_num_threads():
+    """Return the number of OpenMP threads to use in searches
+
+    Initially this is the value returned by omp_get_max_threads(),
+    which is generally 4 unless you set the environment variable
+    OMP_NUM_THREADS to some other value. 
+    
+    It may be any value in the range 1 to get_max_threads(), inclusive.
+    """
+    # I don't want the top-level chemfp module import to import a submodule.
+    import _chemfp
+
+    return _chemfp.get_num_threads()
+
+def set_num_threads(num_threads):
+    """Set the number of OpenMP threads to use in searches
+
+    If `num_threads` is less than one then it is treated as one, and a
+    value greater than get_max_threads() is treated as get_max_threads().
+    """
+    # I don't want the top-level chemfp module import to import a submodule.
+    import _chemfp
+
+    return _chemfp.set_num_threads(num_threads)
+
+def get_max_threads():
+    """Return the maximum number of threads available.
+
+    If OpenMP is not available then this will return 1. Otherwise it
+    returns the maximum number of threads available, as reported by
+    omp_get_num_threads().
+    
+    """
+    # I don't want the top-level chemfp module import to import a submodule.
+    import _chemfp
+
+    return _chemfp.get_max_threads()
+
+
