@@ -3,7 +3,7 @@ import unittest2
 import doctest
 
 def load_tests(loader, tests, ignore):
-    tests.addTests(doctest.DocTestSuite("chemfp.decoders"))
+    tests.addTests(doctest.DocTestSuite("chemfp.encodings"))
     return tests
 
 if __name__ == "__main__":
